from datetime import timedelta
from unittest.mock import Mock, patch

from bravado.exception import HTTPError

from django.test import override_settings
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import NoSocketsTestCase, add_character_to_user, create_fake_user

from ..core import BaseConfig
from ..models import (
    AbstractStandingsRequest,
    CharacterAffiliation,
    Contact,
    ContactSet,
    CorporationDetails,
    FrozenAlt,
    FrozenAuthUser,
    RequestLogEntry,
    StandingRequest,
    StandingRevocation,
)
from .entity_type_ids import CHARACTER_TYPE_ID, CORPORATION_TYPE_ID
from .my_test_data import (
    TEST_STANDINGS_API_CHARID,
    TEST_STANDINGS_API_CHARNAME,
    create_contacts_set,
    create_entity,
    create_standings_char,
    esi_get_alliances_alliance_id_contacts,
    esi_get_alliances_alliance_id_contacts_labels,
    esi_get_corporations_corporation_id,
    esi_post_characters_affiliation,
    load_eve_entities,
)

CORE_PATH = "standingsrequests.core"
MANAGERS_PATH = "standingsrequests.managers"
MODELS_PATH = "standingsrequests.models"
TEST_USER_NAME = "Peter Parker"


class TestContactSetManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_member(TEST_STANDINGS_API_CHARNAME)
        character = create_standings_char()
        add_character_to_user(
            cls.user, character, scopes=["esi-alliances.read_contacts.v1"]
        )
        load_eve_entities()

    @patch(CORE_PATH + ".STANDINGS_API_CHARID", TEST_STANDINGS_API_CHARID)
    @patch(CORE_PATH + ".SR_OPERATION_MODE", "alliance")
    @patch(CORE_PATH + ".SR_OPERATION_MODE", "alliance")
    @patch(MANAGERS_PATH + ".esi")
    def test_can_create_new_from_api(self, mock_esi):
        mock_Contacts = mock_esi.client.Contacts
        mock_Contacts.get_alliances_alliance_id_contacts_labels.side_effect = (
            esi_get_alliances_alliance_id_contacts_labels
        )
        mock_Contacts.get_alliances_alliance_id_contacts.side_effect = (
            esi_get_alliances_alliance_id_contacts
        )

        # labels
        contact_set = ContactSet.objects.create_new_from_api()
        labels = set(contact_set.labels.values_list("label_id", "name"))
        expected = {(1, "blue"), (2, "green"), (3, "yellow"), (4, "red")}
        self.assertSetEqual(labels, expected)

        # all_contacts
        all_contacts = set(
            contact_set.contacts.values_list("eve_entity_id", "standing")
        )
        expected = {
            (1001, 10),
            (1002, 10),
            (1003, 5),
            (1004, 0.01),
            (1005, 0),
            (1006, 0),
            (1008, -5),
            (1009, -10),
            (1010, 5),
            (1110, 5.0),
            (2001, 10.0),
            (2003, 5.0),
            (2102, -10.0),
            (3010, -10.0),
        }
        self.assertSetEqual(all_contacts, expected)

    @patch(CORE_PATH + ".STANDINGS_API_CHARID", TEST_STANDINGS_API_CHARID)
    def test_standings_character_exists(self):
        character = create_standings_char()
        self.assertEqual(BaseConfig.owner_character(), character)

    @patch(CORE_PATH + ".STANDINGS_API_CHARID", TEST_STANDINGS_API_CHARID)
    @patch(MODELS_PATH + ".EveCharacter.objects.create_character")
    def test_standings_character_not_exists(self, mock_create_character):
        character, _ = EveCharacter.objects.get_or_create(
            character_id=TEST_STANDINGS_API_CHARID,
            defaults={
                "character_name": TEST_STANDINGS_API_CHARNAME,
                "corporation_id": 2099,
                "corporation_name": "Dummy Corp",
            },
        )
        mock_create_character.return_value = character
        self.assertEqual(BaseConfig.owner_character(), character)
        self.assertTrue(EveEntity.objects.filter(id=TEST_STANDINGS_API_CHARID).exists())


class TestAbstractStandingsRequestManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_contacts_set()
        cls.user_requestor = AuthUtils.create_member("Bruce Wayne")
        cls.user_manager = AuthUtils.create_user("Mike Manager")

    def test_pending_requests_empty(self):
        self.assertEqual(StandingRequest.objects.pending_requests().count(), 0)

    def test_should_count_pending_requests_correctly(self):
        # given
        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1001,
            contact_type_id=CHARACTER_TYPE_ID,
            is_effective=False,
        )
        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1002,
            contact_type_id=CHARACTER_TYPE_ID,
            is_effective=True,
        )
        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1003,
            contact_type_id=CHARACTER_TYPE_ID,
            is_effective=False,
            action_date=now(),
        )
        # when
        result = StandingRequest.objects.pending_requests()
        # then
        self.assertEqual(result.count(), 1)


@patch(MANAGERS_PATH + ".SR_NOTIFICATIONS_ENABLED", True)
@patch(CORE_PATH + ".STANDINGS_API_CHARID", TEST_STANDINGS_API_CHARID)
@patch(MODELS_PATH + ".SR_STANDING_TIMEOUT_HOURS", 24)
@patch(MANAGERS_PATH + ".notify")
class TestAbstractStandingsRequestProcessRequests(NoSocketsTestCase):
    def setUp(self):
        self.user_manager = AuthUtils.create_user("Mike Manager")
        self.user_requestor = AuthUtils.create_user("Roger Requestor")
        ContactSet.objects.all().delete()
        self.contact_set = create_contacts_set()
        create_standings_char()

    def test_when_pilot_standing_satisfied_in_game_mark_effective_and_inform_user(
        self, mock_notify
    ):
        my_request = StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1002,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
        )
        StandingRequest.objects.process_requests()
        my_request.refresh_from_db()
        self.assertTrue(my_request.is_effective)
        self.assertIsNotNone(my_request.effective_date)
        self.assertEqual(my_request.action_by, self.user_manager)
        self.assertIsNotNone(my_request.action_date)
        self.assertEqual(mock_notify.call_count, 1)
        args, kwargs = mock_notify.call_args
        self.assertEqual(kwargs["user"], self.user_requestor)

    def test_dont_inform_user_when_sr_was_effective_before(self, mock_notify):
        my_request = StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1002,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
            is_effective=True,
            effective_date=now(),
        )
        StandingRequest.objects.process_requests()
        my_request.refresh_from_db()
        self.assertTrue(my_request.is_effective)
        self.assertIsNotNone(my_request.effective_date)
        self.assertEqual(my_request.action_by, self.user_manager)
        self.assertIsNotNone(my_request.action_date)
        self.assertEqual(mock_notify.call_count, 0)

    def test_when_corporation_standing_satisfied_in_game_mark_effective(
        self, mock_notify
    ):
        my_request = StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=2003,
            contact_type_id=CORPORATION_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
        )
        StandingRequest.objects.process_requests()
        my_request.refresh_from_db()
        self.assertTrue(my_request.is_effective)
        self.assertIsNotNone(my_request.effective_date)
        self.assertEqual(my_request.action_by, self.user_manager)
        self.assertIsNotNone(my_request.action_date)
        self.assertTrue(mock_notify.called)

    def test_notify_about_requests_that_are_reset_and_timed_out(self, mock_notify):
        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1008,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now() - timedelta(hours=25),
        )
        StandingRequest.objects.process_requests()
        self.assertEqual(mock_notify.call_count, 2)

    def test_dont_notify_about_requests_that_are_reset_and_not_timed_out(
        self, mock_notify
    ):
        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1008,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now() - timedelta(hours=1),
        )
        StandingRequest.objects.process_requests()
        self.assertEqual(mock_notify.call_count, 0)

    def test_no_action_when_actioned_standing_but_not_in_game_yet(self, mock_notify):
        my_request = StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1002,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
        )
        self.contact_set.contacts.get(eve_entity_id=1002).delete()
        StandingRequest.objects.process_requests()
        my_request.refresh_from_db()
        self.assertFalse(my_request.is_effective)
        self.assertIsNone(my_request.effective_date)
        self.assertEqual(mock_notify.call_count, 0)

    def test_raise_exception_when_called_from_abstract_object(self, mock_notify):
        with self.assertRaises(TypeError):
            AbstractStandingsRequest.objects.process_requests()

    def test_pending_request(self, mock_notify):
        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1001,
            contact_type_id=CHARACTER_TYPE_ID,
            is_effective=False,
        )
        self.assertTrue(AbstractStandingsRequest.objects.has_pending_request(1001))

        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1002,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
            is_effective=True,
            effective_date=now(),
        )
        self.assertFalse(AbstractStandingsRequest.objects.has_pending_request(1002))


class TestAbstractStandingsRequestAnnotations(NoSocketsTestCase):
    def setUp(self):
        self.user_manager = AuthUtils.create_user("Mike Manager")
        self.user_requestor = AuthUtils.create_user("Roger Requestor")
        ContactSet.objects.all().delete()
        self.contact_set = create_contacts_set()
        create_standings_char()

    def test_pending_request_annotation(self):
        # given
        r1 = StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1001,
            contact_type_id=CHARACTER_TYPE_ID,
            is_effective=False,
        )
        r2 = StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1002,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
            is_effective=True,
            effective_date=now(),
        )
        # when
        requests = StandingRequest.objects.all().annotate_is_pending()
        # then
        self.assertTrue(requests.get(pk=r1.pk).is_pending_annotated)
        self.assertFalse(requests.get(pk=r2.pk).is_pending_annotated)


@patch(MODELS_PATH + ".StandingRequest.can_request_corporation_standing")
class TestStandingsRequestValidateRequests(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_contacts_set()
        cls.user = AuthUtils.create_member("Bruce Wayne")

    def setUp(self):
        StandingRequest.objects.all().delete()

    def test_do_nothing_character_request_is_valid(
        self, mock_can_request_corporation_standing
    ):
        AuthUtils.add_permission_to_user_by_name(
            StandingRequest.REQUEST_PERMISSION_NAME, self.user
        )
        request = StandingRequest.objects.get_or_create_2(
            self.user, 1002, StandingRequest.ContactType.CHARACTER
        )

        StandingRequest.objects.validate_requests()
        self.assertTrue(StandingRequest.objects.filter(pk=request.pk).exists())

    def test_create_revocation_if_users_character_has_standing_but_user_no_permission(
        self, mock_can_request_corporation_standing
    ):
        StandingRequest.objects.get_or_create_2(
            self.user, 1002, StandingRequest.ContactType.CHARACTER
        )
        StandingRequest.objects.validate_requests()
        my_revocation = StandingRevocation.objects.get(contact_id=1002)
        self.assertEqual(
            my_revocation.reason, StandingRevocation.Reason.LOST_PERMISSION
        )

    def test_create_revocation_if_users_corporation_is_missing_apis(
        self, mock_can_request_corporation_standing
    ):
        mock_can_request_corporation_standing.return_value = False
        AuthUtils.add_permission_to_user_by_name(
            StandingRequest.REQUEST_PERMISSION_NAME, self.user
        )
        StandingRequest.objects.get_or_create_2(
            self.user, 2001, StandingRequest.ContactType.CORPORATION
        )

        StandingRequest.objects.validate_requests()
        my_revocation = StandingRevocation.objects.get(contact_id=2001)
        self.assertEqual(
            my_revocation.reason, StandingRevocation.Reason.MISSING_CORP_TOKEN
        )

    def test_keep_corp_standing_request_if_all_apis_recorded(
        self, mock_can_request_corporation_standing
    ):
        mock_can_request_corporation_standing.return_value = True
        AuthUtils.add_permission_to_user_by_name(
            StandingRequest.REQUEST_PERMISSION_NAME, self.user
        )
        request = StandingRequest.objects.get_or_create_2(
            self.user, 2001, StandingRequest.ContactType.CORPORATION
        )

        StandingRequest.objects.validate_requests()
        self.assertTrue(StandingRequest.objects.filter(pk=request.pk).exists())


class TestStandingsRequestManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_contacts_set()
        cls.user_requestor = AuthUtils.create_member("Bruce Wayne")
        cls.user_manager = AuthUtils.create_user("Mike Manager")

    def test_should_add_new_request(self):
        # when
        my_request = StandingRequest.objects.get_or_create_2(
            self.user_requestor, 1001, StandingRequest.ContactType.CHARACTER
        )
        # then
        self.assertIsInstance(my_request, StandingRequest)

    def test_should_not_create_new_request_that_already_exists(self):
        # given
        my_request_1 = StandingRequest.objects.get_or_create_2(
            self.user_requestor, 1001, StandingRequest.ContactType.CHARACTER
        )
        # when
        my_request_2 = StandingRequest.objects.get_or_create_2(
            self.user_requestor, 1001, StandingRequest.ContactType.CHARACTER
        )
        # then
        self.assertEqual(my_request_1, my_request_2)


class TestStandingsRevocationManager(NoSocketsTestCase):
    def setUp(self):
        ContactSet.objects.all().delete()
        load_eve_entities()
        my_set = ContactSet.objects.create(name="Dummy Set")
        Contact.objects.create(contact_set=my_set, eve_entity_id=1001, standing=10)
        Contact.objects.create(contact_set=my_set, eve_entity_id=1002, standing=5)
        Contact.objects.create(contact_set=my_set, eve_entity_id=1003, standing=0.01)
        Contact.objects.create(contact_set=my_set, eve_entity_id=1005, standing=0)
        Contact.objects.create(contact_set=my_set, eve_entity_id=1008, standing=-5)
        Contact.objects.create(contact_set=my_set, eve_entity_id=1009, standing=-10)
        self.user_manager = AuthUtils.create_user("Mike Manager")
        self.user_requestor = AuthUtils.create_user("Roger Requestor")

    def test_add_revocation_new(self):
        my_revocation = StandingRevocation.objects.add_revocation(
            1001,
            StandingRevocation.ContactType.CHARACTER,
            user=self.user_requestor,
            reason=StandingRevocation.Reason.OWNER_REQUEST,
        )
        self.assertIsInstance(my_revocation, StandingRevocation)

    def test_add_request_already_exists(self):
        StandingRevocation.objects.add_revocation(
            1001, StandingRevocation.ContactType.CHARACTER
        )
        my_revocation_2 = StandingRevocation.objects.add_revocation(
            1001, StandingRevocation.ContactType.CHARACTER
        )
        self.assertIsNone(my_revocation_2)

    def test_check_standing_satisfied_but_deleted_for_neutral_check_only(self):
        my_revocation = StandingRevocation.objects.add_revocation(
            1999, StandingRevocation.ContactType.CHARACTER
        )
        self.assertTrue(my_revocation.evaluate_effective_standing(check_only=True))

    def test_check_standing_satisfied_but_deleted_for_neutral(self):
        my_revocation = StandingRevocation.objects.add_revocation(
            1999, StandingRevocation.ContactType.CHARACTER
        )
        self.assertTrue(my_revocation.evaluate_effective_standing())
        self.assertTrue(my_revocation.is_effective)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterAffiliationsManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user_manager = AuthUtils.create_user("Mike Manager")
        cls.user_requestor = AuthUtils.create_user("Roger Requestor")

    def test_should_create_new_assocs(self, mock_esi):
        # given
        mock_esi.client.Character.post_characters_affiliation.side_effect = (
            esi_post_characters_affiliation
        )
        create_contacts_set(include_assoc=False)
        StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1002,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
        )
        # when
        CharacterAffiliation.objects.update_from_esi()
        # then
        existing_objects = set(
            CharacterAffiliation.objects.values_list("character_id", flat=True)
        )
        self.assertSetEqual(
            existing_objects, {1001, 1002, 1003, 1004, 1005, 1006, 1008, 1009, 1010}
        )

    def test_should_update_existing_assocs(self, mock_esi):
        # given
        mock_esi.client.Character.post_characters_affiliation.side_effect = (
            esi_post_characters_affiliation
        )
        create_contacts_set(include_assoc=True)
        assoc = CharacterAffiliation.objects.get(character_id=1001)
        assoc.corporation = EveEntity.objects.get(id=2003)
        assoc.save()
        # when
        CharacterAffiliation.objects.update_from_esi()
        # then
        existing_objects = set(
            CharacterAffiliation.objects.values_list("character_id", flat=True)
        )
        self.assertSetEqual(
            existing_objects,
            {1001, 1002, 1003, 1004, 1005, 1006, 1008, 1009, 1010},
        )
        assoc.refresh_from_db()
        self.assertEqual(assoc.corporation_id, 2001)

    def test_should_handle_exception_from_api(self, mock_esi):
        # given
        mock_esi.client.Character.post_characters_affiliation.side_effect = HTTPError(
            Mock()
        )
        create_contacts_set(include_assoc=False)
        # when
        CharacterAffiliation.objects.update_from_esi()

    def test_should_add_new_eve_character_relations(self, mock_esi):
        # given
        create_contacts_set(include_assoc=True)
        eve_character_1001 = create_entity(EveCharacter, 1001)
        # when
        CharacterAffiliation.objects.update_evecharacter_relations()
        # then
        assoc = CharacterAffiliation.objects.get(character_id=1001)
        self.assertEqual(assoc.eve_character, eve_character_1001)

    def test_should_update_existing_eve_character_relations(self, mock_esi):
        # given
        create_contacts_set(include_assoc=True)
        eve_character_1001 = create_entity(EveCharacter, 1001)
        eve_character_1002 = create_entity(EveCharacter, 1002)
        assoc = CharacterAffiliation.objects.get(character_id=1001)
        assoc.eve_character = eve_character_1002
        assoc.save()
        # when
        CharacterAffiliation.objects.update_evecharacter_relations()
        # then
        assoc = CharacterAffiliation.objects.get(character_id=1001)
        self.assertEqual(assoc.eve_character, eve_character_1001)


@patch(MANAGERS_PATH + ".esi")
class TestCorporationDetailsManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eve_entities()

    def test_should_update_corporations(self, mock_esi):
        # given
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        # when
        obj, created = CorporationDetails.objects.update_or_create_from_esi(2001)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.corporation_id, 2001)
        self.assertEqual(obj.alliance_id, 3001)
        self.assertEqual(obj.ceo_id, 1003)
        self.assertEqual(obj.member_count, 3)
        self.assertEqual(obj.ticker, "WYT")
        self.assertIsNone(obj.faction)

    def test_should_not_update_corporations_with_invalid_ceo(self, mock_esi):
        # given
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        # when
        obj, created = CorporationDetails.objects.update_or_create_from_esi(2199)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.corporation_id, 2199)
        self.assertIsNone(obj.ceo_id)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestRequestLogEntryManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eve_entities()
        cls.user_manager = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprices",
        )
        cls.user_requestor = create_fake_user(
            character_id=1002,
            character_name="Peter Parker",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprices",
        )

    def test_should_create_entry_for_confirmed_request(self):
        # given
        my_request = StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=1007,
            contact_type_id=CHARACTER_TYPE_ID,
            action_by=self.user_manager,
            action_date=now(),
        )
        # when
        obj = RequestLogEntry.objects.create_from_standing_request(
            my_request, RequestLogEntry.Action.CONFIRMED, self.user_manager
        )
        # then
        self.assertIsInstance(obj, RequestLogEntry)

    def test_should_create_entry_for_confirmed_revocation(self):
        # given
        my_revocation = StandingRevocation.objects.add_revocation(
            1007,
            StandingRevocation.ContactType.CHARACTER,
            user=self.user_requestor,
            reason=StandingRevocation.Reason.OWNER_REQUEST,
        )
        # when
        obj = RequestLogEntry.objects.create_from_standing_request(
            my_revocation, RequestLogEntry.Action.CONFIRMED, self.user_manager
        )
        # then
        self.assertIsInstance(obj, RequestLogEntry)


class TestFrozenAuthUserManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eve_entities()
        # corporation, _ = EveCorporationInfo.objects.get_or_create(
        #     corporation_id=2001,
        #     defaults={
        #         "corporation_name": "Wayne Technologies",
        #         "corporation_ticker": "WYT",
        #         "member_count": 3,
        #         "ceo_id": 1003,
        #     },
        # )
        cls.member_state = AuthUtils.get_member_state()
        # cls.member_state.member_corporations.add(corporation)

    def test_should_create_full_obj(self):
        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprices",
        )
        user.profile.main_character.faction_id = 500001
        user.profile.main_character.faction_name = "Caldari State"
        user.profile.state = self.member_state
        AuthUtils.disconnect_signals()
        user.profile.main_character.save()
        AuthUtils.connect_signals()
        # when
        obj, created = FrozenAuthUser.objects.get_or_create_from_user(user)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.user, user)
        self.assertEqual(obj.character, EveEntity.objects.get(id=1001))
        self.assertEqual(obj.corporation, EveEntity.objects.get(id=2001))
        self.assertEqual(obj.alliance, EveEntity.objects.get(id=3001))
        self.assertEqual(obj.faction, EveEntity.objects.get(id=500001))
        self.assertEqual(obj.state, self.member_state)

    def test_should_create_obj_without_alliance(self):
        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
        )
        user.profile.main_character.alliance_id = None
        user.profile.main_character.alliance_name = ""
        user.profile.main_character.save()
        # when
        obj, created = FrozenAuthUser.objects.get_or_create_from_user(user)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.user, user)
        self.assertEqual(obj.character, EveEntity.objects.get(id=1001))
        self.assertEqual(obj.corporation, EveEntity.objects.get(id=2001))
        self.assertIsNone(obj.alliance)

    def test_should_create_from_user_without_main(self):
        # given
        user = AuthUtils.create_user("Bruce Wayne")
        # when
        obj, created = FrozenAuthUser.objects.get_or_create_from_user(user)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.user, user)
        self.assertIsNone(obj.character)
        self.assertIsNone(obj.corporation)
        self.assertIsNone(obj.alliance)

    def test_should_not_save_updated_obj(self):
        # given
        user = create_fake_user(character_id=1001, character_name="Bruce Wayne")
        obj, _ = FrozenAuthUser.objects.get_or_create_from_user(user)
        # when
        obj.character_id = 1002
        with self.assertRaises(RuntimeError):
            obj.save()

    def test_should_not_update_obj(self):
        # given
        user = create_fake_user(character_id=1001, character_name="Bruce Wayne")
        obj, _ = FrozenAuthUser.objects.get_or_create_from_user(user)
        # when
        with self.assertRaises(RuntimeError):
            FrozenAuthUser.objects.filter(pk=obj.pk).update(character_id=1002)

    def test_should_get_existing_full_obj(self):
        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprices",
        )
        user.profile.main_character.faction_id = 500001
        user.profile.main_character.faction_name = "Caldari State"
        user.profile.state = self.member_state
        AuthUtils.disconnect_signals()
        user.profile.main_character.save()
        AuthUtils.connect_signals()
        existing_obj, _ = FrozenAuthUser.objects.get_or_create_from_user(user)
        # when
        obj, created = FrozenAuthUser.objects.get_or_create_from_user(user)
        # then
        self.assertFalse(created)
        self.assertEqual(existing_obj, obj)

    def test_should_get_existing_minimal_obj(self):
        # given
        user = AuthUtils.create_user("Bruce Wayne")
        existing_obj, _ = FrozenAuthUser.objects.get_or_create_from_user(user)
        # when
        obj, created = FrozenAuthUser.objects.get_or_create_from_user(user)
        # then
        self.assertFalse(created)
        self.assertEqual(existing_obj, obj)

    def test_should_reset_to_sentinel_when_user_is_deleted(self):
        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprices",
        )
        obj, _ = FrozenAuthUser.objects.get_or_create_from_user(user)
        # when
        user.delete()
        # then
        obj.refresh_from_db()
        self.assertEqual(obj.user.username, "deleted")


class TestFrozenAltManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eve_entities()
        cls.user = create_fake_user(1001, "Bruce Wayne")

    def test_should_create_new_character_without_affiliations(self):
        # given
        my_request = StandingRequest.objects.create(
            user=self.user, contact_id=1002, contact_type_id=CHARACTER_TYPE_ID
        )
        # when
        obj, created = FrozenAlt.objects.get_or_create_from_standing_request(my_request)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.character_id, 1002)
        self.assertIsNone(obj.corporation)
        self.assertIsNone(obj.alliance)
        self.assertEqual(obj.category, FrozenAlt.Category.CHARACTER)

    def test_should_create_new_character_with_affiliations(self):
        # given
        my_request = StandingRequest.objects.create(
            user=self.user, contact_id=1099, contact_type_id=CHARACTER_TYPE_ID
        )
        character = EveEntity.objects.create(
            id=1099, category=EveEntity.CATEGORY_CHARACTER, name="dummy"
        )
        CharacterAffiliation.objects.create(
            character=character,
            corporation_id=2001,
            alliance_id=3001,
            faction_id=500001,
        )
        # when
        obj, created = FrozenAlt.objects.get_or_create_from_standing_request(my_request)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.character_id, 1099)
        self.assertEqual(obj.corporation_id, 2001)
        self.assertEqual(obj.alliance_id, 3001)
        self.assertEqual(obj.faction_id, 500001)
        self.assertEqual(obj.category, FrozenAlt.Category.CHARACTER)

    def test_should_create_new_corporation_without_affiliations(self):
        # given
        my_request = StandingRequest.objects.create(
            user=self.user, contact_id=2099, contact_type_id=CORPORATION_TYPE_ID
        )
        # when
        obj, created = FrozenAlt.objects.get_or_create_from_standing_request(my_request)
        # then
        self.assertTrue(created)
        self.assertIsNone(obj.character)
        self.assertEqual(obj.corporation_id, 2099)
        self.assertIsNone(obj.alliance)
        self.assertEqual(obj.category, FrozenAlt.Category.CORPORATION)

    def test_should_create_new_corporation_with_affiliations(self):
        # given
        my_request = StandingRequest.objects.create(
            user=self.user, contact_id=2099, contact_type_id=CORPORATION_TYPE_ID
        )
        corporation = EveEntity.objects.create(
            id=2099, category=EveEntity.CATEGORY_CORPORATION, name="dummy"
        )
        CorporationDetails.objects.create(
            corporation=corporation,
            alliance_id=3001,
            member_count=99,
            ticker="xyz",
            faction_id=500001,
        )
        # when
        obj, created = FrozenAlt.objects.get_or_create_from_standing_request(my_request)
        # then
        self.assertTrue(created)
        self.assertIsNone(obj.character)
        self.assertEqual(obj.corporation_id, 2099)
        self.assertEqual(obj.alliance_id, 3001)
        self.assertEqual(obj.faction_id, 500001)
        self.assertEqual(obj.category, FrozenAlt.Category.CORPORATION)

    def test_should_get_existing_minimal_obj(self):
        # given
        my_request = StandingRequest.objects.create(
            user=self.user, contact_id=1003, contact_type_id=CHARACTER_TYPE_ID
        )
        existing_obj = FrozenAlt.objects.create(character_id=1003, category="CH")
        # when
        obj, created = FrozenAlt.objects.get_or_create_from_standing_request(my_request)
        # then
        self.assertFalse(created)
        self.assertEqual(existing_obj, obj)

    def test_should_get_existing_full_obj(self):
        # given
        my_request = StandingRequest.objects.create(
            user=self.user, contact_id=1099, contact_type_id=CHARACTER_TYPE_ID
        )
        character = EveEntity.objects.create(
            id=1099, category=EveEntity.CATEGORY_CHARACTER, name="dummy"
        )
        CharacterAffiliation.objects.create(
            character=character,
            corporation_id=2001,
            alliance_id=3001,
            faction_id=500001,
        )
        existing_obj = FrozenAlt.objects.create(
            character_id=1099,
            corporation_id=2001,
            alliance_id=3001,
            category="CH",
            faction_id=500001,
        )
        # when
        obj, created = FrozenAlt.objects.get_or_create_from_standing_request(my_request)
        # then
        self.assertFalse(created)
        self.assertEqual(existing_obj, obj)



from unittest.mock import patch

from django.urls import reverse

from allianceauth.eveonline.models import EveCharacter

from standingsrequests.models import StandingRequest, StandingRevocation

from ..my_test_data import (
    TestViewPagesBase,
    esi_get_corporations_corporation_id,
    esi_post_universe_names,
)

HELPERS_EVECORPORATION_PATH = "standingsrequests.helpers.evecorporation"


@patch(HELPERS_EVECORPORATION_PATH + ".cache")
@patch(HELPERS_EVECORPORATION_PATH + ".esi")
class TestViewManageRequests(TestViewPagesBase):
    def test_request_character(self, mock_esi, mock_cache):
        # given
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        mock_esi.client.Universe.post_universe_names.side_effect = (
            esi_post_universe_names
        )
        mock_cache.get.return_value = None

        alt_id = self.alt_character_1.character_id
        standing_request = StandingRequest.objects.get_or_create_2(
            self.user_requestor,
            alt_id,
            StandingRequest.ContactType.CHARACTER,
        )
        self.client.force_login(self.user_manager)

        # when
        response = self.client.get(reverse("standingsrequests:manage_requests_list"))

        # then
        self.assertEqual(response.status_code, 200)
        data = {obj["contact_id"]: obj for obj in response.context.dicts[3]["requests"]}
        expected = {alt_id}
        self.assertSetEqual(set(data.keys()), expected)
        self.maxDiff = None

        data_alt_1 = data[self.alt_character_1.character_id]
        expected_alt_1 = {
            "contact_id": 1007,
            "contact_name": "James Gordon",
            "contact_icon_url": "https://images.evetech.net/characters/1007/portrait?size=32",
            "corporation_id": 2004,
            "corporation_name": "Metro Police",
            "corporation_ticker": "MP",
            "alliance_id": None,
            "alliance_name": "",
            "has_scopes": True,
            "request_date": standing_request.request_date,
            "action_date": None,
            "state": "Member",
            "main_character_name": "Peter Parker",
            "main_character_ticker": "WYE",
            "main_character_icon_url": "https://images.evetech.net/characters/1002/portrait?size=32",
            "actioned": False,
            "is_effective": False,
            "is_corporation": False,
            "is_character": True,
            "action_by": "(System)",
            "reason": None,
            "labels": [],
        }
        self.assertPartialDictEqual(data_alt_1, expected_alt_1)

    def test_request_corporation(self, mock_esi, mock_cache):
        # given
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        mock_esi.client.Universe.post_universe_names.side_effect = (
            esi_post_universe_names
        )
        mock_cache.get.return_value = None
        alt_id = self.alt_character_1.corporation_id
        standing_request = StandingRequest.objects.get_or_create_2(
            self.user_requestor,
            alt_id,
            StandingRequest.ContactType.CORPORATION,
        )
        self.client.force_login(self.user_manager)

        # when
        response = self.client.get(reverse("standingsrequests:manage_requests_list"))

        # then
        self.assertEqual(response.status_code, 200)
        data = {obj["contact_id"]: obj for obj in response.context.dicts[3]["requests"]}
        expected = {alt_id}
        self.assertSetEqual(set(data.keys()), expected)
        self.maxDiff = None

        expected_alt_1 = {
            "contact_id": 2004,
            "contact_name": "Metro Police",
            "contact_icon_url": "https://images.evetech.net/corporations/2004/logo?size=32",
            "corporation_id": 2004,
            "corporation_name": "Metro Police",
            "corporation_ticker": "MP",
            "alliance_id": None,
            "alliance_name": "",
            "has_scopes": True,
            "request_date": standing_request.request_date,
            "action_date": None,
            "state": "Member",
            "main_character_name": "Peter Parker",
            "main_character_ticker": "WYE",
            "main_character_icon_url": "https://images.evetech.net/characters/1002/portrait?size=32",
            "actioned": False,
            "is_effective": False,
            "is_corporation": True,
            "is_character": False,
            "action_by": "(System)",
            "reason": None,
            "labels": [],
        }
        self.assertPartialDictEqual(data[alt_id], expected_alt_1)


@patch(HELPERS_EVECORPORATION_PATH + ".cache")
@patch(HELPERS_EVECORPORATION_PATH + ".esi")
class TestViewManageRevocations(TestViewPagesBase):
    def test_should_show_character_revocation(self, mock_esi, mock_cache):
        # given
        alt_character = EveCharacter.objects.get(character_id=1110)
        alt_id = alt_character.character_id
        self._create_standing_for_alt(alt_character)
        standing_request = StandingRevocation.objects.add_revocation(
            alt_id,
            StandingRevocation.ContactType.CHARACTER,
            user=self.user_requestor,
            reason=StandingRevocation.Reason.LOST_PERMISSION,
        )
        self.client.force_login(self.user_manager)

        # when
        response = self.client.get(reverse("standingsrequests:manage_revocations_list"))

        # then
        self.assertEqual(response.status_code, 200)
        data = {
            obj["contact_id"]: obj for obj in response.context.dicts[3]["revocations"]
        }
        expected = {alt_id}
        self.assertSetEqual(set(data.keys()), expected)
        self.maxDiff = None

        data_alt_1 = data[alt_id]
        expected_alt_1 = {
            "contact_id": 1110,
            "contact_name": "Phil Coulson",
            "contact_icon_url": "https://images.evetech.net/characters/1110/portrait?size=32",
            "corporation_id": 2110,
            "corporation_name": "Shield",
            "corporation_ticker": "SH",
            "alliance_id": None,
            "alliance_name": "",
            "has_scopes": False,
            "request_date": standing_request.request_date,
            "action_date": None,
            "state": "Member",
            "main_character_name": "Peter Parker",
            "main_character_ticker": "WYE",
            "main_character_icon_url": "https://images.evetech.net/characters/1002/portrait?size=32",
            "actioned": False,
            "is_effective": False,
            "is_corporation": False,
            "is_character": True,
            "action_by": "(System)",
            "reason": "Character owner has lost permission",
            "labels": ["red", "yellow"],
        }
        self.assertPartialDictEqual(data_alt_1, expected_alt_1)

    def test_revoke_corporation(self, mock_esi, mock_cache):
        # given
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        mock_esi.client.Universe.post_universe_names.side_effect = (
            esi_post_universe_names
        )
        mock_cache.get.return_value = None
        alt_id = self.alt_corporation.corporation_id
        self._create_standing_for_alt(self.alt_corporation)
        standing_request = StandingRevocation.objects.add_revocation(
            alt_id,
            StandingRevocation.ContactType.CORPORATION,
            user=self.user_requestor,
        )
        self.client.force_login(self.user_manager)

        # when
        response = self.client.get(reverse("standingsrequests:manage_revocations_list"))

        # then
        self.assertEqual(response.status_code, 200)
        data = {
            obj["contact_id"]: obj for obj in response.context.dicts[3]["revocations"]
        }
        expected = {alt_id}
        self.assertSetEqual(set(data.keys()), expected)
        self.maxDiff = None

        expected_alt_1 = {
            "contact_id": 2004,
            "contact_name": "Metro Police",
            "contact_icon_url": "https://images.evetech.net/corporations/2004/logo?size=32",
            "corporation_id": 2004,
            "corporation_name": "Metro Police",
            "corporation_ticker": "MP",
            "alliance_id": None,
            "alliance_name": "",
            "has_scopes": True,
            "request_date": standing_request.request_date,
            "action_date": None,
            "state": "Member",
            "main_character_name": "Peter Parker",
            "main_character_ticker": "WYE",
            "main_character_icon_url": "https://images.evetech.net/characters/1002/portrait?size=32",
            "actioned": False,
            "is_effective": False,
            "is_corporation": True,
            "is_character": False,
            "action_by": "(System)",
            "reason": "None recorded",
            "labels": [],
        }
        self.assertPartialDictEqual(data[alt_id], expected_alt_1)

    def test_can_show_user_without_main(self, mock_esi, mock_cache):
        # given
        alt_id = self.alt_character_3.character_id
        self._create_standing_for_alt(self.alt_character_3)
        standing_request = StandingRevocation.objects.add_revocation(
            alt_id,
            StandingRevocation.ContactType.CHARACTER,
            user=self.user_former_member,
        )
        self.client.force_login(self.user_manager)

        # when
        response = self.client.get(reverse("standingsrequests:manage_revocations_list"))

        # then
        self.assertEqual(response.status_code, 200)
        data = {
            obj["contact_id"]: obj for obj in response.context.dicts[3]["revocations"]
        }
        expected = {alt_id}
        self.assertSetEqual(set(data.keys()), expected)
        self.maxDiff = None

        data_alt_1 = data[alt_id]
        expected_alt_1 = {
            "contact_id": 1010,
            "contact_name": "Natasha Romanoff",
            "contact_icon_url": "https://images.evetech.net/characters/1010/portrait?size=32",
            "corporation_id": 2102,
            "corporation_name": "Lexcorp",
            "corporation_ticker": "LEX",
            "alliance_id": None,
            "alliance_name": "",
            "has_scopes": False,
            "request_date": standing_request.request_date,
            "action_date": None,
            "state": "Guest",
            "main_character_name": "-",
            "main_character_ticker": "-",
            "main_character_icon_url": "-",
            "actioned": False,
            "is_effective": False,
            "is_corporation": False,
            "is_character": True,
            "action_by": "(System)",
            "reason": "None recorded",
            "labels": ["red"],
        }
        self.assertPartialDictEqual(data_alt_1, expected_alt_1)

    def test_can_handle_requests_without_user(self, mock_esi, mock_cache):
        # setup
        alt_id = 1006
        my_alt = EveCharacter.objects.get(character_id=alt_id)
        self._create_standing_for_alt(my_alt)
        standing_request = StandingRevocation.objects.add_revocation(
            alt_id, StandingRevocation.ContactType.CHARACTER
        )
        self.client.force_login(self.user_manager)

        # when
        response = self.client.get(reverse("standingsrequests:manage_revocations_list"))

        # then
        self.assertEqual(response.status_code, 200)
        data = {
            obj["contact_id"]: obj for obj in response.context.dicts[3]["revocations"]
        }
        expected = {alt_id}
        self.assertSetEqual(set(data.keys()), expected)
        self.maxDiff = None

        data_alt_1 = data[alt_id]
        expected_alt_1 = {
            "contact_id": alt_id,
            "contact_name": "Steven Roger",
            "contact_icon_url": f"https://images.evetech.net/characters/{alt_id}/portrait?size=32",
            "corporation_id": 2003,
            "corporation_name": "CatCo Worldwide Media",
            "corporation_ticker": "CC",
            "alliance_id": None,
            "alliance_name": "",
            "has_scopes": False,
            "request_date": standing_request.request_date,
            "action_date": None,
            "state": "-",
            "main_character_name": "-",
            "main_character_ticker": "-",
            "main_character_icon_url": "-",
            "actioned": False,
            "is_effective": False,
            "is_corporation": False,
            "is_character": True,
            "action_by": "(System)",
            "reason": "None recorded",
            "labels": ["yellow"],
        }
        self.assertPartialDictEqual(data_alt_1, expected_alt_1)



from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.views.decorators.cache import cache_page
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from standingsrequests import __title__
from standingsrequests.app_settings import SR_PAGE_CACHE_SECONDS
from standingsrequests.core import BaseConfig
from standingsrequests.helpers.writers import UnicodeWriter
from standingsrequests.models import ContactSet, ContactType, StandingRequest

from ._common import DEFAULT_ICON_SIZE, add_common_context, label_with_icon

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("standingsrequests.request_standings")
def standings(request):
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contact_set = None
    finally:
        organization = BaseConfig.standings_source_entity()
        last_update = contact_set.date if contact_set else None
    context = {
        "lastUpdate": last_update,
        "organization": organization,
        "show_mains": request.user.has_perm("standingsrequests.view"),
    }
    return render(
        request,
        "standingsrequests/standings.html",
        add_common_context(request, context),
    )


@cache_page(SR_PAGE_CACHE_SECONDS)
@login_required
@permission_required("standingsrequests.request_standings")
def character_standings_data(request):
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()
    character_contacts_qs = (
        contacts.contacts.filter_characters()
        .select_related(
            "eve_entity",
            "eve_entity__character_affiliation",
            "eve_entity__character_affiliation__corporation",
            "eve_entity__character_affiliation__alliance",
            "eve_entity__character_affiliation__faction",
            "eve_entity__character_affiliation__eve_character",
            "eve_entity__character_affiliation__eve_character__character_ownership__user",
            "eve_entity__character_affiliation__eve_character__character_ownership__user__profile__main_character",
            "eve_entity__character_affiliation__eve_character__character_ownership__user__profile__state",
        )
        .prefetch_related("labels")
        .order_by("eve_entity__name")
    )
    characters_data = list()
    for contact in character_contacts_qs:
        character_name_html = label_with_icon(
            contact.eve_entity.icon_url(), contact.eve_entity.name
        )
        if request.user.has_perm("standingsrequests.view"):
            (
                state,
                main_character_name,
                main_character_html,
            ) = _identify_main_for_character(contact)
        else:
            state = main_character_name = main_character_html = ""
        corporation_name, alliance_name, faction_name = _identify_character_assoc(
            contact
        )
        labels = contact.labels_sorted
        characters_data.append(
            {
                "character_id": contact.eve_entity_id,
                "character_name_html": {
                    "display": character_name_html,
                    "sort": contact.eve_entity.name,
                },
                "corporation_name": corporation_name,
                "alliance_name": alliance_name,
                "faction_name": faction_name,
                "standing": contact.standing if contact.standing else "",
                "labels_str": ", ".join(labels),
                "main_character_name": main_character_name,
                "main_character_html": {
                    "display": main_character_html,
                    "sort": main_character_name,
                },
                "state": state,
            }
        )
    return JsonResponse({"data": characters_data})


def _identify_main_for_character(contact):
    try:
        character = contact.eve_entity.character_affiliation.eve_character
        user = character.character_ownership.user
    except (AttributeError, ObjectDoesNotExist):
        state = main_character_name = main_character_html = "-"
    else:
        main = user.profile.main_character
        state = user.profile.state.name if user.profile.state else "-"
        main_character_name = main.character_name
        main_character_ticker = main.corporation_ticker
        main_character_icon_url = main.portrait_url(DEFAULT_ICON_SIZE)
        main_character_html = label_with_icon(
            main_character_icon_url,
            f"[{main_character_ticker}] {main_character_name}",
        )

    return state, main_character_name, main_character_html


def _identify_character_assoc(contact):
    try:
        assoc = contact.eve_entity.character_affiliation
    except (AttributeError, ObjectDoesNotExist):
        corporation_name = "?"
        alliance_name = "?"
        faction_name = "?"
    else:
        corporation_name = assoc.corporation.name
        alliance_name = assoc.alliance.name if assoc.alliance else ""
        faction_name = assoc.faction.name if assoc.faction else ""
    return corporation_name, alliance_name, faction_name


@login_required
@permission_required("standingsrequests.download")
def download_pilot_standings(request):
    logger.info("download_pilot_standings called by %s", request.user)
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = 'attachment; filename="standings.csv"'
    writer = UnicodeWriter(response)
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()
    writer.writerow(
        [
            "character_id",
            "character_name",
            "corporation_id",
            "corporation_name",
            "corporation_ticker",
            "alliance_id",
            "alliance_name",
            "has_scopes",
            "state",
            "main_character_name",
            "main_character_ticker",
            "standing",
            "labels",
        ]
    )

    # lets request make sure all info is there in bulk
    character_contacts = contacts.contacts.all().order_by("eve_entity__name")
    EveEntity.objects.bulk_resolve_names([p.contact_id for p in character_contacts])

    for pilot_standing in character_contacts:
        try:
            char = EveCharacter.objects.get(character_id=pilot_standing.contact_id)
        except EveCharacter.DoesNotExist:
            char = None
        main = ""
        state = ""
        try:
            ownership = CharacterOwnership.objects.get(character=char)
        except CharacterOwnership.DoesNotExist:
            main_character_name = ""
            main = None
        else:
            state = ownership.user.profile.state.name
            main = ownership.user.profile.main_character
            if main is None:
                main_character_name = ""
            else:
                main_character_name = main.character_name
        pilot = [
            pilot_standing.eve_entity_id,
            pilot_standing.eve_entity.name,
            char.corporation_id if char else "",
            char.corporation_name if char else "",
            char.corporation_ticker if char else "",
            char.alliance_id if char else "",
            char.alliance_name if char else "",
            StandingRequest.has_required_scopes_for_request(char),
            state,
            main_character_name,
            main.corporation_ticker if main else "",
            pilot_standing.standing,
            ", ".join([label.name for label in pilot_standing.labels.all()]),
        ]
        writer.writerow([str(v) if v is not None else "" for v in pilot])
    return response


@cache_page(SR_PAGE_CACHE_SECONDS)
@login_required
@permission_required("standingsrequests.request_standings")
def corporation_standings_data(request):
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()
    corporations_qs = (
        contacts.contacts.filter_corporations()
        .select_related(
            "eve_entity",
            "eve_entity__corporation_details",
            "eve_entity__corporation_details__alliance",
            "eve_entity__corporation_details__faction",
        )
        .prefetch_related("labels")
        .order_by("eve_entity__name")
    )
    corporations_data = list()
    standings_requests = {
        obj.contact_id: obj
        for obj in (
            StandingRequest.objects.filter(
                contact_type_id=ContactType.corporation_id
            ).filter(
                contact_id__in=list(
                    corporations_qs.values_list("eve_entity_id", flat=True)
                )
            )
        )
    }
    for contact in corporations_qs:
        alliance_name, faction_name = _identify_corporation_organizations(contact)
        if request.user.has_perm("standingsrequests.view"):
            (
                main_character_name,
                main_character_html,
                state_name,
            ) = _identify_corporation_main(standings_requests, contact)
        else:
            main_character_name = main_character_html = state_name = ""
        labels_str = ", ".join(contact.labels_sorted)
        corporation_html = label_with_icon(
            contact.eve_entity.icon_url(DEFAULT_ICON_SIZE), contact.eve_entity.name
        )
        corporations_data.append(
            {
                "corporation_id": contact.eve_entity_id,
                "corporation_html": {
                    "display": corporation_html,
                    "sort": contact.eve_entity.name,
                },
                "alliance_name": alliance_name,
                "faction_name": faction_name,
                "standing": contact.standing,
                "labels_str": labels_str,
                "state": state_name,
                "main_character_name": main_character_name,
                "main_character_html": {
                    "display": main_character_html,
                    "sort": main_character_name,
                },
            }
        )
    return JsonResponse({"data": corporations_data})


def _identify_corporation_organizations(contact):
    try:
        corporation_details = contact.eve_entity.corporation_details
    except (ObjectDoesNotExist, AttributeError):
        alliance_name = "?"
        faction_name = "?"
    else:
        alliance = corporation_details.alliance
        if alliance:
            alliance_name = alliance.name
        else:
            alliance_name = ""
        faction = corporation_details.faction
        if faction:
            faction_name = faction.name
        else:
            faction_name = ""
    return alliance_name, faction_name


def _identify_corporation_main(standings_requests, contact):
    try:
        standing_request = standings_requests[contact.eve_entity_id]
        user = standing_request.user
        main = user.profile.main_character
    except (KeyError, AttributeError, ObjectDoesNotExist):
        main_character_name = state_name = main_character_html = "-"
    else:
        main_character_name = main.character_name if main else "-"
        main_character_ticker = main.corporation_ticker if main else ""
        main_character_icon_url = main.portrait_url(DEFAULT_ICON_SIZE) if main else "-"
        if main_character_name:
            main_character_html = label_with_icon(
                main_character_icon_url,
                f"[{main_character_ticker}] {main_character_name}",
            )
        else:
            main_character_html = ""
        state_name = user.profile.state.name
    return main_character_name, main_character_html, state_name


@cache_page(SR_PAGE_CACHE_SECONDS)
@login_required
@permission_required("standingsrequests.request_standings")
def alliance_standings_data(request):
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()
    alliances_data = list()
    for contact in (
        contacts.contacts.filter_alliances()
        .select_related("eve_entity")
        .prefetch_related("labels")
        .order_by("eve_entity__name")
    ):
        alliance_html = label_with_icon(
            contact.eve_entity.icon_url(DEFAULT_ICON_SIZE), contact.eve_entity.name
        )
        alliances_data.append(
            {
                "alliance_id": contact.eve_entity_id,
                "alliance_html": {
                    "display": alliance_html,
                    "sort": contact.eve_entity.name,
                },
                "standing": contact.standing,
                "labels_str": ", ".join(contact.labels_sorted),
            }
        )
    return JsonResponse({"data": alliances_data})



from unittest.mock import patch

from eveuniverse.models import EveEntity

from app_utils.testing import NoSocketsTestCase

from ..helpers.evecorporation import EveCorporation
from .my_test_data import esi_get_corporations_corporation_id

MODULE_PATH = "standingsrequests.helpers.evecorporation"


@patch(MODULE_PATH + ".cache")
@patch(MODULE_PATH + ".esi")
class TestEveCorporation(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.corporation = EveCorporation(
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            ticker="WYT",
            ceo_id=1003,
            member_count=3,
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        EveEntity.objects.all().delete()
        EveEntity.objects.create(id=3001, name="Wayne Enterprises", category="alliance")
        cls.maxDiff = None

    def test_init(self, mock_esi, mock_cache):
        self.assertEqual(self.corporation.corporation_id, 2001)
        self.assertEqual(self.corporation.corporation_name, "Wayne Technologies")
        self.assertEqual(self.corporation.ticker, "WYT")
        self.assertEqual(self.corporation.member_count, 3)
        self.assertEqual(self.corporation.alliance_id, 3001)
        self.assertEqual(self.corporation.alliance_name, "Wayne Enterprises")

    def test_str(self, mock_esi, mock_cache):
        expected = "Wayne Technologies"
        self.assertEqual(str(self.corporation), expected)

    def test_get_corp_by_id_not_in_cache(self, mock_esi, mock_cache):
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        expected = self.corporation
        mock_cache.get.return_value = None

        obj = EveCorporation.get_by_id(2001)
        self.assertEqual(obj, expected)
        self.assertTrue(mock_cache.set.called)

    def test_get_corp_by_id_not_in_cache_and_esi_failed(self, mock_esi, mock_cache):
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        mock_cache.get.return_value = None

        obj = EveCorporation.get_by_id(9876)
        self.assertIsNone(obj)

    def test_get_corp_by_id_in_cache(self, mock_esi, mock_cache):
        expected = self.corporation
        mock_cache.get.return_value = expected

        obj = EveCorporation.get_by_id(2001)
        self.assertEqual(obj, expected)

    def test_get_corp_esi(self, mock_esi, mock_cache):
        mock_esi.client.Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        obj = EveCorporation.fetch_corporation_from_api(2102)
        self.assertEqual(obj.corporation_id, 2102)
        self.assertEqual(obj.corporation_name, "Lexcorp")
        self.assertEqual(obj.ticker, "LEX")
        self.assertEqual(obj.member_count, 2)
        self.assertIsNone(obj.alliance_id)

    def test_normal_corp_is_not_npc(self, mock_esi, mock_cache):
        normal_corp = EveCorporation(
            corporation_id=98397665,
            corporation_name="Rancid Rabid Rabis",
            ticker="RANCI",
            member_count=3,
            alliance_id=99005502,
            alliance_name="Same Great Taste",
        )
        self.assertFalse(normal_corp.is_npc)

    def test_npc_corp_is_npc(self, mock_esi, mock_cache):
        normal_corp = EveCorporation(
            corporation_id=1000134,
            corporation_name="Blood Raiders",
            ticker="TBR",
            member_count=22,
        )
        self.assertTrue(normal_corp.is_npc)

    def test_corp_without_members(self, mock_esi, mock_cache):
        normal_corp = EveCorporation(
            corporation_id=98397665,
            corporation_name="Rancid Rabid Rabis",
            ticker="RANCI",
        )
        self.assertIsNone(normal_corp.alliance_name)



from django.db import models
from django.utils.html import format_html

from allianceauth.eveonline.models import EveCharacter

from standingsrequests import __title__
from standingsrequests.constants import DATETIME_FORMAT_HTML
from standingsrequests.core import BaseConfig, ContactType
from standingsrequests.helpers.evecharacter import EveCharacterHelper
from standingsrequests.helpers.evecorporation import EveCorporation
from standingsrequests.models import ContactSet, StandingRequest, StandingRevocation

DEFAULT_ICON_SIZE = 32


def label_with_icon(icon_url: str, text: str):
    return format_html(
        '<span class="text-nowrap">'
        '<img src="{}" class="img-circle" style="width:{}px;height:{}px"> {}'
        "</span>",
        icon_url,
        DEFAULT_ICON_SIZE,
        DEFAULT_ICON_SIZE,
        text,
    )


def add_common_context(request, context: dict) -> dict:
    """adds the common context used by all view"""
    new_context = {
        **{
            "app_title": __title__,
            "operation_mode": str(BaseConfig.operation_mode),
            "pending_total_count": (
                StandingRequest.objects.pending_requests().count()
                + StandingRevocation.objects.pending_requests().count()
            ),
            "DATETIME_FORMAT_HTML": DATETIME_FORMAT_HTML,
        },
        **context,
    }
    return new_context


def compose_standing_requests_data(
    requests_qs: models.QuerySet, quick_check: bool = False
) -> list:
    """composes list of standings requests or revocations based on queryset
    and returns it
    """
    requests_qs = requests_qs.select_related(
        "user", "user__profile__state", "user__profile__main_character"
    )
    # preload data in bulk
    eve_characters = {
        character.character_id: character
        for character in EveCharacter.objects.filter(
            character_id__in=(
                requests_qs.exclude(
                    contact_type_id=ContactType.corporation_id
                ).values_list("contact_id", flat=True)
            )
        )
    }
    # TODO: remove EveCorporation usage
    eve_corporations = {
        corporation.corporation_id: corporation
        for corporation in EveCorporation.get_many_by_id(
            requests_qs.filter(contact_type_id=ContactType.corporation_id).values_list(
                "contact_id", flat=True
            )
        )
    }
    contacts = _identify_contacts(eve_characters, eve_corporations)
    requests_data = list()
    for req in requests_qs:
        (
            main_character_name,
            main_character_ticker,
            main_character_icon_url,
            main_character_html,
            state_name,
        ) = _identify_main(req)

        (
            contact_name,
            contact_icon_url,
            corporation_id,
            corporation_name,
            corporation_ticker,
            alliance_id,
            alliance_name,
            has_scopes,
        ) = _identify_organization(quick_check, eve_characters, eve_corporations, req)

        if contact_name:
            contact_name_html = label_with_icon(contact_icon_url, contact_name)
        else:
            contact_name_html = ""
        if corporation_name:
            organization_html = f"[{corporation_ticker}] {corporation_name}"
            if alliance_name:
                organization_html = format_html(
                    "{}<br>{}", organization_html, alliance_name
                )
        else:
            organization_html = ""

        reason = req.get_reason_display() if req.is_standing_revocation else None
        labels = _fetch_labels(contacts, req)
        requests_data.append(
            {
                "contact_id": req.contact_id,
                "contact_name": contact_name,
                "contact_icon_url": contact_icon_url,
                "contact_name_html": {
                    "display": contact_name_html,
                    "sort": contact_name,
                },
                "corporation_id": corporation_id,
                "corporation_name": corporation_name,
                "corporation_ticker": corporation_ticker,
                "alliance_id": alliance_id,
                "alliance_name": alliance_name,
                "organization_html": organization_html,
                "request_date": req.request_date,
                "action_date": req.action_date,
                "has_scopes": has_scopes,
                "state": state_name,
                "reason": reason,
                "labels": sorted(labels),
                "main_character_name": main_character_name,
                "main_character_ticker": main_character_ticker,
                "main_character_icon_url": main_character_icon_url,
                "main_character_html": main_character_html,
                "actioned": req.is_actioned,
                "is_effective": req.is_effective,
                "is_corporation": req.is_corporation,
                "is_character": req.is_character,
                "action_by": req.action_by.username if req.action_by else "(System)",
            }
        )
    return requests_data


def _identify_contacts(eve_characters, eve_corporations):
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = dict()
    else:
        all_contact_ids = set(eve_characters.keys()) | set(eve_corporations.keys())
        contacts = {
            obj.eve_entity_id: obj
            for obj in contact_set.contacts.prefetch_related("labels").filter(
                eve_entity_id__in=all_contact_ids
            )
        }
    return contacts


def _identify_main(req):
    main_character_name = main_character_ticker = main_character_icon_url = "-"
    main_character_html = "-"
    if req.user:
        state_name = req.user.profile.state.name
        main = req.user.profile.main_character
        if main:
            main_character_name = main.character_name
            main_character_ticker = main.corporation_ticker
            main_character_icon_url = main.portrait_url(DEFAULT_ICON_SIZE)
            main_character_html = label_with_icon(
                main_character_icon_url,
                f"[{main_character_ticker}] {main_character_name}",
            )
    else:
        state_name = "-"
    return (
        main_character_name,
        main_character_ticker,
        main_character_icon_url,
        main_character_html,
        state_name,
    )


def _identify_organization(quick_check, eve_characters, eve_corporations, req):
    if req.is_character:
        if req.contact_id in eve_characters:
            character = eve_characters[req.contact_id]
        else:
            # TODO: remove EveCharacterHelper usage
            character = EveCharacterHelper(req.contact_id)

        contact_name = character.character_name
        contact_icon_url = character.portrait_url(DEFAULT_ICON_SIZE)
        corporation_id = character.corporation_id
        corporation_name = (
            character.corporation_name if character.corporation_name else ""
        )
        corporation_ticker = (
            character.corporation_ticker if character.corporation_ticker else ""
        )
        alliance_id = character.alliance_id
        alliance_name = character.alliance_name if character.alliance_name else ""
        has_scopes = StandingRequest.has_required_scopes_for_request(
            character=character, user=req.user, quick_check=quick_check
        )

    elif req.is_corporation and req.contact_id in eve_corporations:
        corporation = eve_corporations[req.contact_id]
        contact_icon_url = corporation.logo_url(DEFAULT_ICON_SIZE)
        contact_name = corporation.corporation_name
        corporation_id = corporation.corporation_id
        corporation_name = corporation.corporation_name
        corporation_ticker = corporation.ticker
        alliance_id = None
        alliance_name = ""
        has_scopes = not corporation.is_npc and corporation.user_has_all_member_tokens(
            user=req.user, quick_check=quick_check
        )
    else:
        contact_name = ""
        contact_icon_url = ""
        corporation_id = None
        corporation_name = ""
        corporation_ticker = ""
        alliance_id = None
        alliance_name = ""
        has_scopes = False
    return (
        contact_name,
        contact_icon_url,
        corporation_id,
        corporation_name,
        corporation_ticker,
        alliance_id,
        alliance_name,
        has_scopes,
    )


def _fetch_labels(contacts, req):
    try:
        my_contact = contacts[req.contact_id]
    except KeyError:
        labels = []
    else:
        labels = my_contact.labels_sorted
    return labels



# list of entity type IDs used in unit tests

CHARACTER_AMARR_TYPE_ID = 1373
CHARACTER_NI_KUNNI_TYPE_ID = 1374
CHARACTER_CIVRE_TYPE_ID = 1375
CHARACTER_DETEIS_TYPE_ID = 1376
CHARACTER_GALLENTE_TYPE_ID = 1377
CHARACTER_INTAKI_TYPE_ID = 1378
CHARACTER_SEBIESTOR_TYPE_ID = 1379
CHARACTER_BRUTOR_TYPE_ID = 1380
CHARACTER_STATIC_TYPE_ID = 1381
CHARACTER_MODIFIER_TYPE_ID = 1382
CHARACTER_ACHURA_TYPE_ID = 1383
CHARACTER_JIN_MEI_TYPE_ID = 1384
CHARACTER_KHANID_TYPE_ID = 1385
CHARACTER_VHEROKIOR_TYPE_ID = 1386
CHARACTER_DRIFTER_TYPE_ID = 34574

CHARACTER_TYPE_ID = CHARACTER_AMARR_TYPE_ID

CORPORATION_TYPE_ID = 2

ALLIANCE_TYPE_ID = 16159

FACTION_CALDARI_STATE_TYPE_ID = 500001

FACTION_TYPE_ID = FACTION_CALDARI_STATE_TYPE_ID



from unittest.mock import Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from esi.models import CallbackRedirect, Token

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import _generate_token, _store_as_Token, generate_invalid_pk

from ..decorators import token_required_by_state
from .my_test_data import create_eve_objects

MODULE_PATH = "standingsrequests.decorators"
PATH_MODELS = "standingsrequests.models"


@patch(PATH_MODELS + ".StandingRequest.get_required_scopes_for_state")
@patch(MODULE_PATH + ".select_token", spec=True)
@patch(MODULE_PATH + ".sso_redirect", spec=True)
@patch(MODULE_PATH + "._check_callback")
class TestTokenRequiredByState(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_eve_objects()
        cls.user = AuthUtils.create_member("Bruce Wayne")
        cls.token = _store_as_Token(
            _generate_token(
                character_id=1002, character_name=cls.user.username, scopes=["abc"]
            ),
            cls.user,
        )
        cls.factory = RequestFactory()

    def setUp(self):
        CallbackRedirect.objects.all().delete()

    def generate_get_request(self):
        request = self.factory.get("https://www.example.com/my_view/")
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        request.session.save()
        return request

    def generate_post_request(self, data=dict(), user=None):
        request = self.factory.post("https://www.example.com/my_view/", data)
        request.user = self.user if not user else user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        request.session.save()
        return request

    def test_can_present_user_with_token_choices_when_token_matches(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=False)
        def my_view(request, tokens):
            return tokens

        mock_check_callback.return_value = None
        mock_get_required_scopes_for_state.return_value = ["abc"]

        response = my_view(self.generate_get_request())
        self.assertTrue(response, mock_select_token())

    def test_prompt_user_to_add_new_token_when_token_matches_not(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=False)
        def my_view(request, tokens):
            return tokens

        mock_check_callback.return_value = None
        mock_get_required_scopes_for_state.return_value = ["xyz"]

        response = my_view(self.generate_get_request())
        self.assertTrue(response, mock_sso_redirect())

    def test_proceeds_to_view_when_token_is_received(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=True)
        def my_view(request, token):
            return token

        mock_check_callback.return_value = self.token
        mock_get_required_scopes_for_state.return_value = ["abc"]

        response = my_view(self.generate_get_request())
        self.assertEqual(response, self.token)

    def test_user_has_selected_to_add_new_token(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=True)
        def my_view(request, token):
            return token

        mock_check_callback.return_value = None
        mock_get_required_scopes_for_state.return_value = ["abc"]

        data = {"_add": True}
        response = my_view(self.generate_post_request(data))

        self.assertEqual(response, mock_sso_redirect())

    def test_user_has_provided_a_new_token(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=True)
        def my_view(request, token):
            return token

        mock_check_callback.return_value = None
        mock_get_required_scopes_for_state.return_value = ["abc"]

        data = {"_token": self.token.pk}
        response = my_view(self.generate_post_request(data))

        self.assertEqual(response, self.token)

    def test_return_to_selection_if_provided_token_from_user_has_vanished(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=True)
        def my_view(request, token):
            return token

        mock_check_callback.return_value = None
        mock_get_required_scopes_for_state.return_value = ["abc"]

        data = {"_token": generate_invalid_pk(Token)}
        response = my_view(self.generate_post_request(data))

        self.assertEqual(response, mock_sso_redirect())

    def test_prompt_user_to_add_new_token_if_provided_has_wrong_scope(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=True)
        def my_view(request, token):
            return token

        mock_check_callback.return_value = None
        mock_get_required_scopes_for_state.return_value = ["xyz"]

        data = {"_token": self.token.pk}
        response = my_view(self.generate_post_request(data))

        self.assertEqual(response, mock_sso_redirect())

    def test_prompt_user_to_add_new_token_if_token_provided_with_wrong_user(
        self,
        mock_check_callback,
        mock_sso_redirect,
        mock_select_token,
        mock_get_required_scopes_for_state,
    ):
        @token_required_by_state(new=True)
        def my_view(request, token):
            return token

        mock_check_callback.return_value = None
        mock_get_required_scopes_for_state.return_value = ["xyz"]

        data = {"_token": self.token.pk}
        other_user = AuthUtils.create_user("Lex Luther")
        response = my_view(self.generate_post_request(data, user=other_user))

        self.assertEqual(response, mock_sso_redirect())



import json

from django.http import JsonResponse

from app_utils.testing import NoSocketsTestCase, response_text


class PartialDictEqualMixin:
    def assertPartialDictEqual(self, d1: dict, d2: dict):
        """Assert that d1 equals d2 for the subset of keys of d1."""
        subset = {k: v for k, v in d1.items() if k in d2}
        self.assertDictEqual(subset, d2)


class NoSocketsTestCasePlus(PartialDictEqualMixin, NoSocketsTestCase):
    pass


def json_response_to_python_2(response: JsonResponse, data_key="data") -> object:
    """Convert JSON response into Python object."""
    data = json.loads(response_text(response))
    return data[data_key]


def json_response_to_dict_2(response: JsonResponse, key="id", data_key="data") -> dict:
    """Convert JSON response into dict by given key."""
    return {x[key]: x for x in json_response_to_python_2(response, data_key)}
